{$A+,B-,D+,E-,F+,G+,I-,L+,N+,O-,R-,S-,V-}

UNIT Lights;

INTERFACE

USES Extragraph,Maths,Basis,Messages;

CONST
   (*** Verschiedene Lichtmodelle ***)

      LightParallel=0;
      LightCentral=1;
      LightSpot=2;
      LightStick=3;
      LightDisk=4;

      maxLight=10;


TYPE LightRec=RECORD
         kind:BYTE;  (* 0:parallel *)
                     (* 1:central  *)
                     (* 2:Spot     *)
                     (* 3:Stab     *)
                     (* 4:Scheibe  *)
         c:Vector3;               (* Zentrum der Lichtquelle, bei Stab Anfang *)
         d:Vector3;               (* bei Stab: Vektor zum Endpunkt            *)
         sina,cosa,sinb,cosb,sinc,cosc:RealType;
                                  (* bei Scheibe: Verdrehung der X-Y-Scheibe  *)
         r:RealType;              (* bei Scheibe: Radius                      *)
         l:Vector3;               (* Bei Parallel und Spot: Richtung          *)
         index:RealType;          (* Bei Spot: Index des Kegels               *)
         darkness:RealType;       (* Entfernungsindex                         *)
         f:RealType;              (* Anteil dieser Quelle am Gesamtbild (0..1)*)
         LightColor:RGBfloat;     (* Farbe des Lichtes                        *)
     END;


VAR
    LightCnt:INTEGER;
    Light:ARRAY[1..maxLight] OF LightRec;


PROCEDURE SetLightNormal(ll:Vector3;ff:RealType;col:RGBColor);
(*** Setzt Parameter fr paralleles Licht.    ***)
(***  llx,lly,llz:Richtung des Lichts         ***)
(***  ff:Anteil des Lichts am Gesamtbild      ***)
(***  col:Farbe des Lichts                    ***)


PROCEDURE SetLightCentral(cc:Vector3;darkIndex,ff:RealType;col:RGBColor);
(*** Setzt Parameter fr zentrales Licht.     ***)
(***  ccx,ccy,ccz:Position des Lichts         ***)
(***  darkIndex:Abschwchung bei Entfernung:  ***)
(***       0:keine..unendlich:stark           ***)
(***  ff:Anteil des Lichts am Gesamtbild      ***)
(***  col:Farbe des Lichts                    ***)


PROCEDURE SetLightStick(cc1,cc2:Vector3;darkIndex,ff:RealType;col:RGBColor);
(*** Setzt Parameter fr stabfrmiges Licht.  ***)
(***  ccx1,ccy1,ccz1:Anfangspunkt des Stabes  ***)
(***  ccx2,ccy2,ccz2:Endpunkt des Stabes      ***)
(***  darkIndex:Abschwchung bei Entfernung:  ***)
(***       0:keine..unendlich:stark           ***)
(***  ff:Anteil des Lichts am Gesamtbild      ***)
(***  col:Farbe des Lichts                    ***)
(***  Achtung: Bezglich Schattenwurf wird der***)
(***   Stabeffekt nur bei smoothed=TRUE       ***)
(***   sichtbar, sonst wird der Schatten bzgl ***)
(***   Mittelpunkt berechnet                  ***)


PROCEDURE SetLightDisk(cc:Vector3;anga,angb,angc,rr,darkIndex,ff:RealType;col:RGBColor);
(*** Setzt Parameter fr scheibenfrmiges     ***)
(*** Licht (in x-y-Ebene)                     ***)
(***  ccx,ccy,ccz:Mittelpunkt der Scheibe     ***)
(***  a,b,c:Winkel, mit denen die Scheibe ge- ***)
(***         dreht wird z.B.                  ***)
(***          0, 0, 0: x-y-Ebene              ***)
(***          0,90, 0: x-z-Ebene              ***)
(***  r:Radius                                ***)
(***  darkIndex:Abschwchung bei Entfernung:  ***)
(***       0:keine..unendlich:stark           ***)
(***  ff:Anteil des Lichts am Gesamtbild      ***)
(***  col:Farbe des Lichts                    ***)
(***  Achtung: Bezglich Schattenwurf wird der***)
(***   Scheibeneffekt nur bei smoothed=TRUE   ***)
(***   sichtbar, sonst wird der Schatten bzgl ***)
(***   Mittelpunkt berechnet                  ***)


PROCEDURE SetLightSpot(cc,ll:Vector3;brightIndex,darkIndex,ff:RealType;col:RGBColor);
(*** Setzt Parameter fr Spot-Licht.          ***)
(***  ccx,ccy,ccz:Position des Lichts         ***)
(***  llx,lly,llz:Richtung                    ***)
(***  brightIndex:Wie stark schwcht das Licht***)
(***       am Rand ab 1:garnicht..undendlich  ***)
(***  darkIndex:Abschwchung bei Entfernung:  ***)
(***       0:keine..unendlich:stark           ***)
(***  ff:Anteil des Lichts am Gesamtbild      ***)
(***  col:Farbe des Lichts                    ***)


PROCEDURE GetLight(SourceInd:INTEGER;Scan:BOOLEAN;ppp:RealType;VAR p,n:Vector3;
                   VAR li:Vector3;VAR maxL,intensity,ff:RealType;VAR col:RGBfloat);
(*** Gibt die Richtung und Intensitt des     ***)
(*** Lichtes in einem bestimmten Punktes zu-  ***)
(*** rck.                                    ***)
(*** Eingabe:                                 ***)
(***   SourceInd:Nr der Lichtquelle           ***)
(***   Scan: Bei Stab:findet Intervall-Scan   ***)
(***     statt?                               ***)
(***   p:Position des gefragten Punkts        ***)
(***   n:Richtung auf die getestet wird       ***)
(***                                          ***)
(*** Rckgabe:                                ***)
(***   li:Richtung zum Licht                  ***)
(***   maxL:Entfernung zur Lichtquelle        ***)
(***   intensity:Welche Kraft hat das Licht   ***)
(***      fr diesen Punkt (0..1)             ***)
(***   ff:Anteilswert                         ***)
(***   col:Farbe                              ***)


PROCEDURE GetLightColor(VAR dd,r:Vector3;minVal:RealType;
                        VAR mainShaddow:BOOLEAN;VAR ddd:RealType);
(* Gibt den Wert einer beleuchteten Flche zurck    ***)

PROCEDURE GetBrightColor(VAR li,r:Vector3;brightness,brightIndex:RealType;
                         VAR brightCol,col:RGBfloat;VAR newcol:RGBfloat);
(* Gibt den Glanz einer beleuchteten Flche zurck    ***)


IMPLEMENTATION

CONST Epsilon=0.00001;


PROCEDURE SetLightNormal(ll:Vector3;ff:RealType;col:RGBColor);
VAR l:RealType;
BEGIN
    INC(LightCnt);
    WITH Light[LightCnt] DO
    BEGIN
        kind:=LightParallel;
        l.x:=ll.x;
        l.y:=ll.y;
        l.z:=ll.z;
        f:=ff;
        copyRGBcol(col,LightColor);
        VnormSelf(l)
    END
END;


PROCEDURE SetLightCentral(cc:Vector3;darkIndex,ff:RealType;col:RGBColor);
BEGIN
    INC(LightCnt);
    WITH Light[LightCnt] DO
    BEGIN
        kind:=LightCentral;
        c.x:=cc.x;
        c.y:=cc.y;
        c.z:=cc.z;
        f:=ff;
        copyRGBcol(col,LightColor);
        darkness:=darkIndex;
    END
END;


PROCEDURE SetLightStick(cc1,cc2:Vector3;darkIndex,ff:RealType;col:RGBColor);
BEGIN
    INC(LightCnt);
    WITH Light[LightCnt] DO
    BEGIN
        kind:=LightStick;
        c.x:=cc1.x;
        c.y:=cc1.y;
        c.z:=cc1.z;
        d.x:=cc2.x-cc1.x;
        d.y:=cc2.y-cc1.y;
        d.z:=cc2.z-cc1.z;
        f:=ff;
        copyRGBcol(col,LightColor);
        darkness:=darkIndex;
    END
END;


PROCEDURE SetLightDisk(cc:Vector3;anga,angb,angc,rr,darkIndex,ff:RealType;col:RGBColor);
BEGIN
    INC(LightCnt);
    WITH Light[LightCnt] DO
    BEGIN
        kind:=LightDisk;
        c.x:=cc.x;
        c.y:=cc.y;
        c.z:=cc.z;
        anga:=anga*PiDiv180;
        angb:=angb*PiDiv180;
        angc:=angc*PiDiv180;
        sina:=Sin(anga);
        cosa:=Cos(anga);
        sinb:=Sin(angb);
        cosb:=Cos(angb);
        sinc:=Sin(angc);
        cosc:=Cos(angc);
        r:=rr;
        f:=ff;
        copyRGBcol(col,LightColor);
        darkness:=darkIndex;
    END
END;


PROCEDURE SetLightSpot(cc,ll:Vector3;brightIndex,darkIndex,ff:RealType;col:RGBColor);
BEGIN
    INC(LightCnt);
    WITH Light[LightCnt] DO
    BEGIN
        kind:=LightSpot;
        c.x:=cc.x;
        c.y:=cc.y;
        c.z:=cc.z;
        l.x:=ll.x;
        l.y:=ll.y;
        l.z:=ll.z;
        index:=brightIndex;
        darkness:=darkIndex;
        f:=ff;
        copyRGBcol(col,LightColor);
        VnormSelf(l)
    END
END;



PROCEDURE GetLight(SourceInd:INTEGER;Scan:BOOLEAN;ppp:RealType;VAR p,n:Vector3;
                   VAR li:Vector3;VAR maxL,intensity,ff:RealType;VAR col:RGBfloat);
VAR ss,nnn,kkk:RealType;
    k:Vector3;
    err:BOOLEAN;
    rr:Vector3;
    pp:RealType;
BEGIN
    WITH Light[SourceInd] DO
    BEGIN
        ff:=f;
        col:=LightColor;
        CASE kind OF
            LightParallel:
             BEGIN
                 li:=l;
                 maxL:=infinite;
                 intensity:=1;
              END;
            LightCentral:
             BEGIN
                 VVsub(p,c,li);
                 maxL:=Vsize(li);
                 IF maxL<epsilon THEN
                 BEGIN
                     li.x:=0;
                     li.y:=-1;
                     li.z:=0
                 END
                 ELSE 
                     Vresize(li,1/maxL);
                 intensity:=1/(0.5+maxL*darkness);
                 IF intensity>1 THEN
                     intensity:=1;
             END;
            LightSpot:
             BEGIN
                 VVsub(p,c,li);
                 maxL:=Vsize(li);
                 IF maxL<epsilon THEN
                 BEGIN
                     li.x:=0;
                     li.y:=-1;
                     li.z:=0
                 END
                 ELSE
                     Vresize(li,1/maxL);
                 ss:=VVsp(l,li);
                 IF (ss<epsilon) THEN
                     ss:=epsilon;
                 intensity:=1/(0.5+maxL*darkness);
                 IF intensity>1 THEN
                     intensity:=1;
                 intensity:=(Power(ss,index))*intensity;
             END;
            LightStick:
             BEGIN
                 IF NOT Scan THEN
                 BEGIN
                     VVcross(d,n,k);
                     lgs(n.x,k.x,d.x,p.x-c.x, 
                         n.y,k.y,d.y,p.y-c.y,
                         n.z,k.z,d.z,p.z-c.z,
                         nnn,kkk,ppp,
                         err);
                     IF err THEN
                         ppp:=0.5
                     ELSE IF ppp<0 THEN
                         ppp:=0
                     ELSE IF ppp>1 THEN
                         ppp:=1;
                 END;
                 li.x:=p.x-(c.x+d.x*ppp);
                 li.y:=p.y-(c.y+d.y*ppp);
                 li.z:=p.z-(c.z+d.z*ppp);
                 maxL:=Vsize(li);
                 IF maxL<epsilon THEN
                 BEGIN
                     li.x:=0;
                     li.y:=-1;
                     li.z:=0
                 END
                 ELSE
                     Vresize(li,1/maxL);
                 intensity:=1/(0.5+maxL*darkness);
                 IF intensity>1 THEN
                     intensity:=1;
             END;
            LightDisk:
             BEGIN
                 IF Scan THEN
                 BEGIN
                     rr.x:=1;
                     rr.y:=0;
                     rr.z:=0;
                     turn2D(rr.x,rr.y,0,0,Power(ppp+0.01,0.3+0.3)*6*Pi);
                     pp:=r;
                     Turn3DRight0(rr.x,rr.y,rr.z,sina,cosa,sinb,cosb,sinc,cosc);
                     li.x:=p.x-(c.x+rr.x*pp);  
                     li.y:=p.y-(c.y+rr.y*pp);
                     li.z:=p.z-(c.z+rr.z*pp)
                 END
                 ELSE
                     VVsub(p,c,li);
                 maxL:=Vsize(li);
                 IF maxL<epsilon THEN
                 BEGIN
                     li.x:=0;
                     li.y:=-1;
                     li.z:=0
                 END
                 ELSE
                     Vresize(li,1/maxL);
                 intensity:=1/(0.5+maxL*darkness);
                 IF intensity>1 THEN
                     intensity:=1;
             END;
        END
    END
END;


PROCEDURE GetLightColor(VAR dd,r:Vector3;minVal:RealType;
                        VAR mainShaddow:BOOLEAN;VAR ddd:RealType);


VAR l:RealType;
    ss:RealType;
BEGIN
    ss:=VVsp(r,dd);
    mainShaddow:=(ss>=0);
    ddd:=(1+ss)*(minVal-1)+1;
    IF ddd<=minVal THEN
        ddd:=minVal
END;


PROCEDURE GetBrightColor(VAR li,r:Vector3;brightness,brightIndex:RealType;
                        VAR brightCol,col:RGBfloat;VAR newcol:RGBfloat);

VAR sdd:RealType;
    dd:RealType;

BEGIN
    VnormSelf(r);
    dd:=-VVsp(r,li);
    IF dd<epsilon THEN
        sdd:=0
    ELSE
        sdd:=Power(dd,brightIndex)*brightness;

    RGBpartition(col,brightCol,newcol,sdd);
END;




BEGIN
    LightCnt:=0
END.
