{$A+,B-,D+,E-,F+,G+,I-,L+,N+,O-,R-,S-,V-}
{$DEFINE PROFILERx}
{$DEFINE IntNamesx}
{$DEFINE Tracex}
{$DEFINE ObjNamesx}


UNIT SolidObjects;

INTERFACE

USES Maths,Messages,Basis,ObjPoints,Intervalls,Textures
{$IFDEF PROFILER}
     ,Profiler
{$ENDIF}
     ;

TYPE

     PSolidObject=^SolidObject;
     SolidObject=OBJECT
         c:Vector3;
         m:Matrix3;
         mInverse:Matrix3;
         f1,f2,f3,f4,f5,f6,f4half,f5half,f6half,fr:RealType;

         smoothed:BOOLEAN;

         isReflection:BOOLEAN;
         reflectionValue:PScalarTexture;

         isRefraction:BOOLEAN;
         refractionValue:PScalarTexture;
         refractionIndex:RealType;

         isShaddowed:BOOLEAN;
         shaddowed:RealType;
         selfshaddow:RealType;

         isBright:BOOLEAN;
         brightness:RealType;
         brightIndex:RealType;

         isTransparent:BOOLEAN;
         transparencyValue:PScalarTexture;

         shaddowesOthers:BOOLEAN;

         texture:PVectorTexture;

{$IFDEF ObjNames}
         name:STRING20;
{$ENDIF}

         PROCEDURE GetIntervall(VAR p,r:Vector3;VAR resultIntervall:IntervallSet);
    END;

IMPLEMENTATION



PROCEDURE SolidObject.GetIntervall(VAR p,r:Vector3;VAR resultIntervall:IntervallSet);
CONST ep=0.000001;
VAR p2,r2:Vector3;
    aD,bD,cD,aDRez:RealType;
    cnt:WORD;
    ll1,ll2:RealType;
    nn1,nn2:Vector3;
    phalf,w:RealType;
    i:BYTE;
{$IFDEF Trace}
    wStr:STRING[10];
    illegal:BOOLEAN;
{$ENDIF}
    f1p2x,f2p2y,f3p2z,f1r2x,f2r2y,f3r2z:RealType;
BEGIN
{$IFDEF PROFILER}
    BeginIt(5);
{$ENDIF}

    VVsub(p,c,p2);
    MVmulSelf(m,p2);
    MVMul(m,r,r2);

    f1p2x:=f1*p2.x;
    f2p2y:=f2*p2.y;
    f3p2z:=f3*p2.z;
    f1r2x:=f1*r2.x;
    f2r2y:=f2*r2.y;
    f3r2z:=f3*r2.z;

    aD:=f1r2x*r2.x+f2r2y*r2.y+f3r2z*r2.z;
    bD:=2*(f1r2x*p2.x+f2r2y*p2.y+f3r2z*p2.z+f4half*r2.x+f5half*r2.y+f6half*r2.z);
    cD:=f1p2x*p2.x+f2p2y*p2.y+f3p2z*p2.z+f4*p2.x+f5*p2.y+f6*p2.z-fr;
    IF (ABS(aD)>ep) THEN
    BEGIN
        aDRez:=1/aD;
        phalf:=bD*aDRez*c05;
        w:=SQR(phalf)-cD*aDRez;
        IF (w<=0) THEN
        BEGIN
            (* Fall I*)
            cnt:=0;
{$IFDEF Trace}
            illegal:=FALSE;
            wStr:='Case I'
{$ENDIF}
        END
        ELSE BEGIN
            (* Fall II*)
            cnt:=2;
            w:=Sqrt(w);
            ll1:=-phalf-w;
            ll2:=-phalf+w
        END
    END
    ELSE BEGIN
         IF (ABS(bD)>ep) THEN
         BEGIN
             (* Fall III*)
             cnt:=1;
             ll1:=-cD/bD
         END
         ELSE BEGIN
             IF (ABS(cD)>ep) THEN
             BEGIN
                 (* Fall IV*)
                 cnt:=0;
{$IFDEF Trace}
                 illegal:=TRUE;
                 wStr:='Case IV'
{$ENDIF}
             END
             ELSE BEGIN
                 (* Fall V*)
                 cnt:=0;
{$IFDEF Trace}
                 illegal:=TRUE;
                 wStr:='Case V'
{$ENDIF}
             END
         END
    END;
    IF cnt=0 THEN
    BEGIN
        resultIntervall.cnt:=0;
{$IFDEF Trace}
        IF illegal THEN
            trace.Add('SolidObject:'+Name+'->Illegal solution '+wStr);
{$ENDIF}
    END
    ELSE IF cnt=1 THEN
    BEGIN
        nn1.x:=f1p2x+f1r2x*ll1+f4half;
        nn1.y:=f2p2y+f2r2y*ll1+f5half;
        nn1.z:=f3p2z+f3r2z*ll1+f6half;
        MVmulSelf(mInverse,nn1);
        resultIntervall.cnt:=1;
        IF VVspLess0(nn1,r) THEN
            WITH resultIntervall.block[1] DO
            BEGIN
                WITH iFrom DO
                BEGIN
                    l:=ll1;
                    n:=nn1;
                    VnormSelf(n);
                    obj:=@self
                END;

                iTo:=infinitePoint
            END
        ELSE
            WITH resultIntervall.block[1] DO
            BEGIN
                iFrom:=ninfinitePoint;

                WITH iTo DO
                BEGIN
                    l:=ll1;
                    n:=nn1;
                    VnormSelf(n);
                    obj:=@self
                END
            END
    END
    ELSE BEGIN
        nn1.x:=f1p2x+f1r2x*ll1+f4half;
        nn1.y:=f2p2y+f2r2y*ll1+f5half;
        nn1.z:=f3p2z+f3r2z*ll1+f6half;
        MVmulSelf(mInverse,nn1);
        nn2.x:=f1p2x+f1r2x*ll2+f4half;
        nn2.y:=f2p2y+f2r2y*ll2+f5half;
        nn2.z:=f3p2z+f3r2z*ll2+f6half;
        MVmulSelf(mInverse,nn2);
        IF VVspLess0(nn1,r) THEN
        BEGIN
            resultIntervall.cnt:=1;
            WITH resultIntervall.block[1] DO
            BEGIN
                WITH iFrom DO
                BEGIN
                    l:=ll1;
                    n:=nn1;
                    VnormSelf(n);
                    obj:=@self
                END;
                WITH iTo DO
                BEGIN
                    l:=ll2;
                    n:=nn2;
                    VnormSelf(n);
                    obj:=@self
                END
            END
        END
        ELSE BEGIN
            resultIntervall.cnt:=2;
            WITH resultIntervall.block[1] DO
            BEGIN
                iFrom:=ninfinitePoint;
                WITH iTo DO
                BEGIN
                    l:=ll1;
                    n:=nn1;
                    VnormSelf(n);
                    obj:=@self
                END
            END;
            WITH resultIntervall.block[2] DO
            BEGIN
                WITH iFrom DO
                BEGIN
                    l:=ll2;
                    n:=nn2;
                    VnormSelf(n);
                    obj:=@self
                END;
                iTo:=infinitePoint
            END
        END
    END;
{$IFDEF IntNames}
    resultIntervall.name:=name
{$ENDIF}
{$IFDEF PROFILER}
    ;EndIt(5);
{$ENDIF}
END;


END.
