{$A+,B-,D-,E-,F+,G+,I-,L-,N-,O-,R-,S-,V-}


UNIT VM;

INTERFACE

CONST maxP=63;


TYPE ByteArray=ARRAY[0..65000] OF BYTE;
     VirtualMemory=RECORD
         cntx,cnty:LONGINT;
         size:WORD;
         dx,dy:WORD;
         np:WORD;
         lastP:WORD;
         swapDir:STRING;
         lastUseID:LONGINT;
         pageSize:WORD;
         page:ARRAY[0..maxP] OF RECORD
             useID:LONGINT;
             changed:BOOLEAN;
             realx,realy:LONGINT;
             data:^ ByteArray
         END
     END;


FUNCTION VMResult:INTEGER;
(*** Gibt bei den Fehlercode der letzen Operation ***)
(*** zurck.                                      ***)


PROCEDURE ReWriteVM(VAR virtMem:VirtualMemory;cntx,cnty:LONGINT;size:WORD;Fill:POINTER;
                    dx,dy,np:WORD;SwapDir:STRING);
(*** Legt einen neuen virtuellen Speicherraum an. ***)
(*** SwapDir legt fest, wo auf der Platte der Aus-***)
(*** lagerungsbereich liegt. Dieses Directory mu ***)
(*** schon angelegt sein. x,y ist die Gre des   ***)
(*** Arrays [0..x-1,0..y-1]. size ist die Gre   ***)
(*** eines Elementes. Fill ist die Initialbeset-  ***)
(*** zung. dx,dy gibt an welche Kachelindizes zu- ***)
(*** sammengefat werden. np ist die Anzahl der   ***)
(*** Kacheln.                                     ***)


PROCEDURE ResetVM(VAR virtMem:VirtualMemory;SwapDir:STRING);
(*** Ermglicht den Zugriff auf einen virtuellen  ***)
(*** Speicherbereich.                             ***)


PROCEDURE ResetVMP(VAR virtMem:VirtualMemory;SwapDir:STRING;pages:WORD);
(*** Ermglicht den Zugriff auf einen virtuellen  ***)
(*** Speicherbereich. pages gibt die Zahl der     ***)
(*** phys. Seiten an (in nderung der Def.).      ***)


PROCEDURE CloseVM(VAR virtMem:VirtualMemory);
(*** Schreibt alle im Speicher befindlichen Ka-   ***)
(*** cheln auf Platte.                            ***)


PROCEDURE FlushVM(VAR virtMem:VirtualMemory);
(*** Schreibt alle temp. Daten auf Disk, schliet ***)
(*** aber nicht den virtuellen Speicher.          ***)


PROCEDURE GetVM(VAR virtMem:VirtualMemory;x,y:LONGINT;value:POINTER);
(*** Gibt den Wert des Feldelements x,y aus.      ***)


PROCEDURE PutVM(VAR virtMem:VirtualMemory;x,y:LONGINT;value:POINTER);
(*** Schreibt den Wert eines Feldelementes.       ***)


IMPLEMENTATION

VAR VMError:INTEGER;

FUNCTION VMResult:INTEGER;
BEGIN
    VMResult:=VMError;
    VMError:=0
END;


FUNCTION nextID(VAR virtMem:VirtualMemory):LONGINT;
BEGIN
    INC(virtMem.lastUseID);
    nextID:=virtMem.lastUseID
END;


PROCEDURE PageName(VAR virtMem:VirtualMemory;VAR name:STRING;x,y:LONGINT);
VAR part:STRING;

    PROCEDURE PartName(VAR part:STRING;x:LONGINT);
    BEGIN
        part:='';
        WHILE x>0 DO
        BEGIN
            part:=part+CHR(65+(x MOD 26));
            x:=x DIV 26;
        END;
        IF part='' THEN
            part:='A'
    END;

BEGIN
    PartName(name,x);
    PartName(part,y);
    name:=virtMem.swapDir+'\'+name+part+'.PAG'
END;


PROCEDURE ReWriteVM(VAR virtMem:VirtualMemory;cntx,cnty:LONGINT;size:WORD;Fill:POINTER;
                    dx,dy,np:WORD;SwapDir:STRING);
VAR i:WORD;
    name:STRING;
    f:FILE;
    x,y:LONGINT;
BEGIN
    virtMem.cntx:=cntx;
    virtMem.cnty:=cnty;
    virtMem.size:=size;
    virtMem.dx:=dx;
    virtMem.dy:=dy;
    virtMem.np:=np;
    virtMem.lastp:=PRED(np);
    virtMem.swapDir:=swapDir;
    virtMem.lastUseID:=0;
    IF LONGINT(size)*dx*dy>65000 THEN
    BEGIN
        WriteLn('PageSize>64k');
        Halt(1)
    END;
    virtMem.pageSize:=size*dx*dy;
    WITH virtMem DO
        FOR i:=0 TO lastp DO
            WITH page[i] DO
            BEGIN
                useID:=nextID(virtMem);
                changed:=FALSE;
                realx:=-1;
                realy:=-1;
                IF MaxAvail<pageSize THEN
                BEGIN
                    VMError:=32767;
                    Exit
                END;
                GetMem(data,pageSize)
            END;
    WITH virtMem,page[0] DO
    BEGIN
        FOR x:=0 TO PRED(dx) DO
            FOR y:=0 TO PRED(dy) DO
                Move(Fill^,data^[(x*dy+y)*size],size);
        FOR x:=0 TO PRED(cntx) DIV dx DO
            FOR y:=0 TO PRED(cnty) DIV dy DO
            BEGIN
                PageName(virtMem,name,x,y);
                Assign(f,name);
                ReWrite(f,1);
                VMError:=IOResult;
                IF VMError=0 THEN
                BEGIN
                    BlockWrite(f,data^,pagesize);
                    VMError:=IOResult
                END;
                IF VMError=0 THEN
                BEGIN
                    Close(f);
                    VMError:=IOResult
                END
            END
    END;
    IF VMError=0 THEN
    BEGIN
        Assign(f,swapDir+'\VMHEAD.DAT');
        ReWrite(f,1);
        VMError:=IOResult
    END;
    IF VMError=0 THEN
    BEGIN
        BlockWrite(f,virtMem,SizeOf(virtMem));
        VMError:=IOResult
    END;
    IF VMError=0 THEN
    BEGIN
        Close(f);
        VMError:=IOResult
    END
END;




PROCEDURE ResetVM(VAR virtMem:VirtualMemory;SwapDir:STRING);
VAR i:WORD;
    f:FILE;
BEGIN
    Assign(f,swapDir+'\VMHEAD.DAT');
    Reset(f,1);
    VMError:=IOResult;
    IF VMError=0 THEN
    BEGIN
        BlockRead(f,virtMem,SizeOf(virtMem));
        VMError:=IOResult
    END;
    IF VMError=0 THEN
    BEGIN
        Close(f);
        VMError:=IOResult
    END;
    IF VMError=0 THEN
        WITH virtMem DO
        BEGIN
            lastUseID:=0;
            FOR i:=0 TO lastp DO
                WITH page[i] DO
                BEGIN
                    useID:=nextID(virtMem);
                    changed:=FALSE;
                    realx:=-1;
                    realy:=-1;
                    IF MaxAvail<pageSize THEN
                    BEGIN
                        VMError:=32767;
                        Exit
                    END;
                    GetMem(data,pageSize)
                END
        END
END;



PROCEDURE ResetVMP(VAR virtMem:VirtualMemory;SwapDir:STRING;pages:WORD);
VAR i:WORD;
    f:FILE;
BEGIN
    Assign(f,swapDir+'\VMHEAD.DAT');
    Reset(f,1);
    VMError:=IOResult;
    IF VMError=0 THEN
    BEGIN
        BlockRead(f,virtMem,SizeOf(virtMem));
        VMError:=IOResult
    END;
    IF VMError=0 THEN
    BEGIN
        Close(f);
        VMError:=IOResult
    END;
    IF VMError=0 THEN
        WITH virtMem DO
        BEGIN
            lastp:=PRED(pages);
            lastUseID:=0;
            FOR i:=0 TO lastp DO
                WITH page[i] DO
                BEGIN
                    useID:=nextID(virtMem);
                    changed:=FALSE;
                    realx:=-1;
                    realy:=-1;
                    IF MaxAvail<pageSize THEN
                    BEGIN
                        VMError:=32767;
                        Exit
                    END;
                    GetMem(data,pageSize)
                END
        END
END;



PROCEDURE CloseVM(VAR virtMem:VirtualMemory);
VAR i:WORD;
BEGIN
    FlushVM(virtMem);
    WITH virtMem DO
        FOR i:=0 TO lastp DO
            WITH page[i] DO
                FreeMem(data,pageSize)
END;


PROCEDURE FlushPage(VAR virtMem:VirtualMemory;i:WORD);
VAR f:FILE;
    name:STRING;
BEGIN
    WITH virtMem DO
        WITH page[i] DO
            BEGIN
                PageName(virtMem,name,realx,realy);
                Assign(f,name);
                ReWrite(f,1);
                VMError:=IOResult;
                IF VMError=0 THEN
                BEGIN
                    BlockWrite(f,data^,pagesize);
                    VMError:=IOResult
                END;
                IF VMError=0 THEN
                BEGIN
                    Close(f);
                    VMError:=IOResult
                END
            END
END;


PROCEDURE LoadPage(VAR virtMem:VirtualMemory;i:WORD;xx,yy:LONGINT);
VAR f:FILE;
    name:STRING;
BEGIN
    WITH virtMem DO
        WITH page[i] DO
            BEGIN
                PageName(virtMem,name,xx,yy);
                Assign(f,name);
                Reset(f,1);
                VMError:=IOResult;
                IF VMError=0 THEN
                BEGIN
                    BlockRead(f,data^,pagesize);
                    VMError:=IOResult
                END;
                IF VMError=0 THEN
                BEGIN
                    Close(f);
                    VMError:=IOResult
                END;
                changed:=FALSE;
                realx:=xx;
                realy:=yy
            END
END;


PROCEDURE FlushVM(VAR virtMem:VirtualMemory);
VAR i:WORD;
BEGIN
    WITH virtMem DO
        FOR i:=0 TO lastp DO
        BEGIN
            FlushPage(virtMem,i);
            IF VMError<>0 THEN
                Exit
        END
END;


PROCEDURE GetPage(VAR virtMem:VirtualMemory;x,y:LONGINT;VAR i,sub:WORD);
VAR xx,yy:LONGINT;
    lowUseID:LONGINT;
    ii,lowi:WORD;
BEGIN
    WITH virtMem DO
    BEGIN
        xx:=x DIV dx;
        yy:=y DIV dy;
        sub:=((x MOD dx)*dy+(y MOD dy))*size;
        lowUseID:=2000000000;
        FOR ii:=0 TO lastp DO
            WITH page[ii] DO
            BEGIN
                IF (xx=realx) AND (yy=realy) THEN
                BEGIN
                    i:=ii;
                    Exit
                END;
                IF useID<lowUseID THEN
                BEGIN
                    lowUseID:=useID;
                    lowi:=ii
                END
            END;
        i:=lowi;
        WITH page[i] DO
        BEGIN
            IF changed THEN
                FlushPage(virtMem,i);
            IF VMError=0 THEN
                LoadPage(virtMem,i,xx,yy)
        END;
    END
END;


PROCEDURE GetVM(VAR virtMem:VirtualMemory;x,y:LONGINT;value:POINTER);
VAR i:WORD;
    sub:WORD;
BEGIN
    GetPage(virtMem,x,y,i,sub);
    IF VMError=0 THEN
        WITH virtMem,page[i] DO
        BEGIN
            Move(data^[sub],value^,size);
            useID:=nextID(virtMem)
        END
END;


PROCEDURE PutVM(VAR virtMem:VirtualMemory;x,y:LONGINT;value:POINTER);
VAR i:WORD;
    sub:WORD;
BEGIN
    GetPage(virtMem,x,y,i,sub);
    IF VMError=0 THEN
        WITH virtMem,page[i] DO
        BEGIN
            Move(value^,data^[sub],size);
            useID:=nextID(virtMem);
            changed:=TRUE
        END
END;



BEGIN
    VMError:=0
END.